from collections.abc import AsyncIterator

import pytest

from aiterio import Component, ComponentError


class Add(Component):
    def __init__(self, addend: int = 1) -> None:
        super().__init__()
        self._addend = addend

    async def process(self, item: int) -> AsyncIterator[int]:
        yield item + self._addend


class Multiply(Component):
    def __init__(self, factor: int = 1) -> None:
        super().__init__()
        self._factor = factor

    async def process(self, item: int) -> AsyncIterator[int]:
        yield item * self._factor


async def test_processing_integers() -> None:
    component = Add(1).source(range(5))
    result = [value async for value in component]
    assert result == [1, 2, 3, 4, 5]


async def test_multiple_components() -> None:
    pipeline = Add(1).source(range(5)).then(Multiply(factor=2))
    result = [value async for value in pipeline]
    assert result == [2, 4, 6, 8, 10]


async def test_non_iterator_source() -> None:
    component = Add(1).source(1234)  # type: ignore[arg-type]
    with pytest.raises(ComponentError):
        await component.run()


async def test_type() -> None:
    Component._type_checking = True  # noqa: SLF001
    component = Add(1).source(["a", "b", "c"])
    with pytest.raises(TypeError):
        await component.run()
    Component._type_checking = False  # noqa: SLF001


async def test_source_last() -> None:
    pipeline = Add(2).then(Multiply(3)).source([1, 2, 3])
    result = [value async for value in pipeline]
    assert result == [9, 12, 15]


async def test_source_multiple() -> None:
    pipeline = Add(2).then(Multiply(3))
    result = [value async for value in pipeline.source([1, 2, 3])]
    assert result == [9, 12, 15]

    result = [value async for value in pipeline.source([3, 2, 1])]
    assert result == [15, 12, 9]


async def test_source_exhaustion() -> None:
    pipeline = Add(2).source([1, 2, 3]).then(Multiply(3))
    result = [value async for value in pipeline]
    assert result == [9, 12, 15]

    # source is not called again so this is exhausted
    result = [value async for value in pipeline]
    assert result == []

    result = [value async for value in pipeline.source([3, 2, 1])]
    assert result == [15, 12, 9]



import asyncio
import timeit
from collections.abc import AsyncIterator, Iterator

from aiterio import Component


# 1. Synchronous function to increment a number
def inc(x: int) -> int:
    return x + 1


# 2. Iterator version
def iter_inc() -> Iterator[int]:
    for i in range(1000):
        yield inc(i)


# 3. Asynchronous iterator version
async def async_iter_inc() -> AsyncIterator[int]:
    for i in range(1000):
        await asyncio.sleep(0)  # necessary for async behavior
        yield inc(i)


# 4. Using Component
class IncComponent(Component):
    async def process(self, item: int) -> AsyncIterator[int]:
        yield inc(item)


async def run_sync() -> None:
    list(map(inc, range(1000)))


async def run_iter_inc() -> None:
    list(iter_inc())


async def run_async_iter() -> None:
    async for _ in async_iter_inc():
        pass


async def run_async_pipeline() -> None:
    component = IncComponent().source(async_iter_inc())
    async for _ in component:
        pass


async def run_async_pipeline_type_checking() -> None:
    component = IncComponent().source(async_iter_inc())
    async for _ in component:
        pass


# Benchmarking functions
def benchmark_sync() -> None:
    asyncio.run(run_sync())


def benchmark_iter() -> None:
    asyncio.run(run_iter_inc())


def benchmark_async_iter() -> None:
    asyncio.run(run_async_iter())


def benchmark_async_pipeline() -> None:
    asyncio.run(run_async_pipeline())


def benchmark_async_pipeline_type_checking() -> None:
    asyncio.run(run_async_pipeline_type_checking())


# Running benchmarks
if __name__ == "__main__":
    print("Benchmarking synchronous function:")
    print(timeit.timeit("benchmark_sync()", globals=globals(), number=100))

    print("Benchmarking iterator:")
    print(timeit.timeit("benchmark_iter()", globals=globals(), number=100))

    print("Benchmarking asynchronous iterator:")
    print(timeit.timeit("benchmark_async_iter()", globals=globals(), number=100))

    print("Benchmarking Component:")
    print(timeit.timeit("benchmark_async_pipeline()", globals=globals(), number=100))

    print("Benchmarking Component with type_checking:")
    print(timeit.timeit("benchmark_async_pipeline_type_checking()", globals=globals(), number=100))



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Iterable
from typing import Any, ClassVar
from weakref import ReferenceType


class ComponentError(Exception):
    """Raised when a component cannot process correctly"""


class Component(ABC, AsyncIterator):
    """
    A component that can be passed to a pipeline to create a chain of commands.
    Consuming an Iterable and yielding an AsyncIterator.
    It can also be called directly by iterating over its process method with individual values.
    """

    _instances: ClassVar[list[ReferenceType[Component]]] = []

    @classmethod
    def instances(cls) -> list[Component]:
        """
        Conveniency class method returning a list of all running instances of Component
        Also remove instances that are not referenced anymore from _instances
        Can be used for graceful shutdown
        """
        return [instance for reference in cls._instances if (instance := reference()) is not None]

    @classmethod
    def register_instance(cls, self: Component) -> None:
        cls._instances.append(ReferenceType(self, cls.remove_instance))

    @classmethod
    def remove_instance(cls, ref: ReferenceType) -> None:
        """
        Remove an instance from the list of tracked instances.
        """
        cls._instances.remove(ref)

    def __init__(self) -> None:
        """
        Init registers the instance. By default an instance does nothing.
        """
        self.__class__.register_instance(self)
        self._prev_component: Component | None = None
        self._source_iter: AsyncIterator
        self._is_running = True

    @property
    def is_running(self) -> bool:
        """
        Can be overriden to define what is_running means.
        """
        return self._is_running

    def stop(self) -> None:
        """
        Can be overriden for custom stop logic.
        """
        self._is_running = False
        if self._prev_component:
            self._prev_component.stop()

    @staticmethod
    async def _wrap_source(source: Source) -> AsyncIterator[Any]:
        """
        A helper method to wrap synchronous iterators with async functionality and yield items.
        """
        if isinstance(source, AsyncIterator):
            async for item in source:
                yield item
        elif isinstance(source, Iterable):
            for item in source:
                yield item
                await asyncio.sleep(0)  # Yield control to keep the event loop responsive
        else:
            raise ComponentError("Source data must be an Iterable or AsyncIterator.")

    async def _make_source_iter(self, source: Source) -> AsyncIterator[Any]:
        """
        Generate items from the source data and process each item.
        This method combines fetching and processing logic.
        """
        async for item in self._wrap_source(source):
            async for processed_item in self.process(item):
                yield processed_item
                await asyncio.sleep(0)  # Yield control to keep the event loop responsive

    def source(self, source: Source) -> Component:
        """
        This method can be called on any element of a chain.
        It transforms a Source into an AsyncIterator and based on the type of Source:
        - Iterator/AsyncIterator: it will propagate to the first component and reset the pipeline.
        - Component: such as when called by `.then` it sets the previous component as the source.
        """
        self._source_iter = None

        if self._prev_component:
            self._prev_component.source(source)
            self._source_iter = self._make_source_iter(self._prev_component)
            return self

        if isinstance(source, Component):
            self._prev_component = source

        self._source_iter = self._make_source_iter(source)
        return self

    def __aiter__(self) -> AsyncIterator[Any]:
        return self

    async def __anext__(self) -> Any:
        if not self.is_running:
            raise StopAsyncIteration
        return await self._source_iter.__anext__()

    @abstractmethod
    async def process(self, item: Any) -> AsyncIterator[Any]:
        """
        Method to be overridden by subclasses to define specific processing logic.
        """
        yield item

    def then(self, component: Component) -> Component:
        """
        Chain another component to process the output of this component.
        """
        component.source(self)
        return component

    async def run(self) -> None:
        """
        Consume all items in this component.
        Can be overriden for custom logic.
        """
        async for _ in self:
            pass


Source = Component | Iterable[Any] | AsyncIterator[Any]
